# Copyright (C) 2001-2010 Python Software Foundation; All Rights Reserved.
# Copyright (C) 2011 Dan Pascu; See LICENSE for details.
#

from __future__ import absolute_import

__all__ = ['OrderedDict']

from itertools import imap as imap
from operator import eq as eq
from weakref import ref

try:
    from _abcoll import Mapping, KeysView, ValuesView, ItemsView
except ImportError:
    Mapping = dict
    def KeysView(obj):
        return obj.keys()
    def ValuesView(obj):
        return obj.values()
    def ItemsView(obj):
        return obj.items()

try:
    from thread import get_ident
except ImportError:
    from dummy_thread import get_ident


def recursive_repr(user_function):
    """Decorator to make a repr function return "..." for a recursive call"""
    repr_running = set()

    def wrapper(self):
        key = id(self), get_ident()
        if key in repr_running:
            return '...'
        repr_running.add(key)
        try:
            result = user_function(self)
        finally:
            repr_running.discard(key)
        return result

    # Can't use functools.wraps() here because of bootstrap issues
    wrapper.__module__ = getattr(user_function, '__module__')
    wrapper.__doc__ = getattr(user_function, '__doc__')
    wrapper.__name__ = getattr(user_function, '__name__')
    return wrapper


class Node(object):
    __slots__ = ('key', '_prev_ref', '_next_ref', '__weakref__')

    def __init__(self, key=None, prev=None, next=None):
        self.key = key
        self.prev = prev
        self.next = next

    def __repr__(self):
        return "Node(key=%r, prev=Node(key=%r), next=Node(key=%r))" % (self.key, self.prev.key, self.next.key)

    def _get_prev(self):
        return self._prev_ref if self._prev_ref is None else self._prev_ref()
    def _set_prev(self, value):
        self._prev_ref = value if value is None else ref(value)

    def _get_next(self):
        return self._next_ref if self._next_ref is None else self._next_ref()
    def _set_next(self, value):
        self._next_ref = value if value is None else ref(value)

    prev = property(_get_prev, _set_prev)
    next = property(_get_next, _set_next)

    del _get_prev, _set_prev, _get_next, _set_next


# This is a modified version of OrderedDict from collections.py in python 2.7
#
class OrderedDict(dict):
    'Dictionary that remembers insertion order'
    # An inherited dict maps keys to values.
    # The inherited dict provides __getitem__, __len__, __contains__, and get.
    # The remaining methods are order-aware.
    # Big-O running times for all methods are the same as for regular dictionaries.

    # The internal self.__map dictionary maps keys to nodes in a doubly linked list.
    # The circular doubly linked list starts and ends with a sentinel node.
    # The sentinel node never gets deleted (this simplifies the algorithm).

    def __init__(self, *args, **kwds):
        """
        Initialize an ordered dictionary.  Signature is the same as for
        regular dictionaries, but keyword arguments are not recommended
        because their insertion order is arbitrary.
        """
        if len(args) > 1:
            raise TypeError('expected at most 1 positional argument, got %d' % len(args))
        try:
            self.__root
        except AttributeError:
            self.__root = Node()   # sentinel node
            self.__root.prev = self.__root.next = self.__root
            self.__map = {}
        OrderedDict.update(self, *args, **kwds)

    def __setitem__(self, key, value, dict_setitem=dict.__setitem__):
        """od.__setitem__(i, y) <==> od[i]=y"""
        # Setting a new item creates a new node which goes at the end of the linked
        # list, and the inherited dictionary is updated with the new key/value pair.
        if key not in self:
            root = self.__root
            last = root.prev
            last.next = root.prev = self.__map[key] = Node(key, prev=last, next=root)
        dict_setitem(self, key, value)

    def __delitem__(self, key, dict_delitem=dict.__delitem__):
        """od.__delitem__(y) <==> del od[y]"""
        # Deleting an existing item uses self.__map to find the node which is
        # then removed by updating the links in the predecessor and successor nodes.
        dict_delitem(self, key)
        node = self.__map.pop(key)
        node.prev.next = node.next
        node.next.prev = node.prev

    def __iter__(self):
        """od.__iter__() <==> iter(od)"""
        # Traverse the linked list in order.
        root = self.__root
        node = root.next
        while node is not root:
            yield node.key
            node = node.next

    def __reversed__(self):
        """od.__reversed__() <==> reversed(od)"""
        # Traverse the linked list in reverse order.
        root = self.__root
        node = root.prev
        while node is not root:
            yield node.key
            node = node.prev

    @recursive_repr
    def __repr__(self):
        """od.__repr__() <==> repr(od)"""
        if not self:
            return '%s()' % (self.__class__.__name__,)
        return '%s(%r)' % (self.__class__.__name__, self.items())

    def __reduce__(self):
        """Return state information for pickling"""
        items = [[k, self[k]] for k in self]
        inst_dict = vars(self).copy()
        inst_dict.pop('_OrderedDict__root')
        inst_dict.pop('_OrderedDict__map')
        if inst_dict:
            return (self.__class__, (items,), inst_dict)
        return self.__class__, (items,)

    def __eq__(self, other):
        """
        od.__eq__(y) <==> od==y.  Comparison to another OD is order-sensitive
        while comparison to a regular mapping is order-insensitive.
        """
        if isinstance(other, OrderedDict):
            return len(self)==len(other) and all(imap(eq, self.iteritems(), other.iteritems()))
        return dict.__eq__(self, other)

    def __ne__(self, other):
        return not (self == other)

    @classmethod
    def fromkeys(cls, iterable, value=None):
        """
        od.fromkeys(S[, v]) -> New ordered dictionary with keys from S
        and values equal to v (which defaults to None).
        """
        d = cls()
        for key in iterable:
            d[key] = value
        return d

    def clear(self):
        """od.clear() -> None.  Remove all items from od."""
        self.__root.prev = self.__root.next = self.__root
        self.__map.clear()
        dict.clear(self)

    def copy(self):
        """od.copy() -> a shallow copy of od"""
        return self.__class__(self)

    def setdefault(self, key, default=None):
        try:
            return self[key]
        except KeyError:
            self[key] = default
        return default

    def update(self, *args, **kwds):
        if len(args) > 1:
            raise TypeError("update() takes at most 1 positional argument (%d given)" % len(args))

        other = args[0] if args else ()

        if isinstance(other, Mapping):
            for key in other:
                self[key] = other[key]
        elif hasattr(other, "keys"):
            for key in other.keys():
                self[key] = other[key]
        else:
            for key, value in other:
                self[key] = value
        for key, value in kwds.items():
            self[key] = value

    __marker = object()

    def pop(self, key, default=__marker):
        try:
            value = self[key]
        except KeyError:
            if default is self.__marker:
                raise
            return default
        else:
            del self[key]
            return value

    def popitem(self, last=True):
        """
        od.popitem() -> (k, v), return and remove a (key, value) pair.
        Pairs are returned in LIFO order if last is true or FIFO order if false.
        """
        if not self:
            raise KeyError('dictionary is empty')
        key = next(reversed(self) if last else iter(self))
        value = self.pop(key)
        return key, value

    def keys(self):
        return list(self)

    def values(self):
        return [self[key] for key in self]

    def items(self):
        return [(key, self[key]) for key in self]

    def iterkeys(self):
        return iter(self)

    def itervalues(self):
        for key in self:
            yield self[key]

    def iteritems(self):
        for key in self:
            yield (key, self[key])

    def viewkeys(self):
        """od.viewkeys() -> a set-like object providing a view on od's keys"""
        return KeysView(self)

    def viewvalues(self):
        """od.viewvalues() -> an object providing a view on od's values"""
        return ValuesView(self)

    def viewitems(self):
        """od.viewitems() -> a set-like object providing a view on od's items"""
        return ItemsView(self)


